/*
Accediendo directamente al linux file system ext2  usando libreria ext2fs
Este programa solo funciona con file system de tipo ext2.
Si Ud. no tiene una particion de tipo ext2 puede armarse una de la siguiente forma:

1. crear un archivo de imagen de file system de 500K (por ejemplo, 500 bloques de 1K cada uno):

$ dd if=/dev/zero of=my.fs.image bs=1024 seek=500 count=0

2. crear un file system de tipo ext2 en la imagen my.fs.image:

$ /sbin/mke2fs -t ext2 ./my.fs.image

3. montar la imagen de file system en, por ejemplo /tmp/fs:

$ mkdir /tmp/fs
$ mount -o loop=/dev/loop0 my.fs.image /tmp/fs

4. compruebe que el file system esta montado:

$ mount

5. copie cualquier archivo, cree carpetas, etc. en /tmp/fs

6. utilice ls para saber con que i-nodo se encuentra asociado cada archivo (el primero numero de la izquierda):

$ ls -li /tmp/fs

7. Ejecute este programa indicando el file system y el numero de i-nodo que pretende consultar. Ejemplo:

$ sudo mkdir /tmp/fs
$ sudo mount -o loop=/dev/loop0 my.fs.image /tmp/fs
$ sudo mount
/dev/hda1 on / type ext3 (rw,errors=remount-ro)
tmpfs on /lib/init/rw type tmpfs (rw,nosuid,mode=0755)
proc on /proc type proc (rw,noexec,nosuid,nodev)
sysfs on /sys type sysfs (rw,noexec,nosuid,nodev)
procbususb on /proc/bus/usb type usbfs (rw)
udev on /dev type tmpfs (rw,mode=0755)
tmpfs on /dev/shm type tmpfs (rw,nosuid,nodev)
devpts on /dev/pts type devpts (rw,noexec,nosuid,gid=5,mode=620)
/home/grchere/gccwork/src/sockets/my.fs.image on /tmp/fs type ext2 (rw,loop=/dev/loop0)
$ ls -li /tmp/fs
total 51
12 -rwxr-xr-x 1 grchere grchere 37721 2010-08-18 11:31 dbf.c
11 drwx------ 2 root    root    12288 2010-08-18 11:15 lost+found
13 -rw-r--r-- 1 grchere grchere     2 2010-08-18 19:00 pepe
$ 

$ ./tp102 /home/grchere/gccwork/src/sockets/my.fs.image 12

$



UTN FRD SO
Guillermo Cherencio

#include <ext2fs/ext2_fs.h>
#include <ext2fs/ext2_err.h>
#include <ext2fs/ext2_ext_attr.h>
#include <ext2fs/ext2fs.h>
#include <ext2fs/ext2_io.h>
#include <ext2fs/ext3_extents.h>
#include <ext2fs/tdb.h>

Para compilar este programa:
   1. Se requiere previamente tener instalado el paquete ext2fs-dev ( $ sudo aptitude install ext2fs-dev )
   2. En su sistema deberian estar los headers que requiere este programa y la libreria estatica libext2fs.a
   
gcc -Wall -o fstest fstest.c -lext2fs
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <ext2fs/ext2_fs.h>
#include <ext2fs/ext2fs.h>

// prototipo de funciones locales
int leo_bloques_de_datos(ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private);
errcode_t leo_bloque(ext2_filsys fs,blk_t blocknr,char *block_buffer);
void imprimo_bloque(char *block_buffer,int blocksize);
void imprimo_bloque_indireccion(char *block_buffer,int blocksize);
void imprimo_indireccion(short indireccion,ext2_filsys fs,blk_t blocknr,short modo_interactivo);

// estructura utilizada para comunicar con funciones de callback
typedef struct {
   int contador;
   int total_bloques;
   short modo_interactivo;
} fs_data;

int main(int argc, char * argv[]) {
   if ( argc != 3 && argc != 4) { 
      printf("\nModo de uso: $ tp102 <sistema de archivos> <numero inodo> [-ni]\n-ni  modo No Interactivo\n");
      printf("Ejemplo:\n$ ./tp102 /dev/hda1 146409\nUtilice el comando mount para saber los montajes actuales:\n$ mount\n");
      exit(1); 
   }
   printf("main():inicio\n");
   short modo_interactivo=1; // asumo modo interactivo
   ext2_filsys fs = NULL; // esto en realidad es un puntero! hay un typedef dentro de ext2fs.h
   ext2_ino_t ino =  atoi(argv[2]); 
   if ( argc == 4 ) if ( strcmp(argv[3],"-ni") == 0 ) modo_interactivo=0;
   struct ext2_inode inode;
   memset(&inode,0,sizeof(struct ext2_inode));
   int superblock = 0, blocksize = 0, openflags = EXT2_FLAG_RW;
   char *filesystem = argv[1]; //"/home/grchere/gccwork/src/sockets/my.fs.image";  // "/dev/hda1";
   //printf("main():ext2fs_open\n");
   errcode_t ecode = ext2fs_open (filesystem, openflags, superblock, blocksize,unix_io_manager, &fs);
   if ( ecode ) {
      printf("main():Error, ext2fs_open() retorna: %ld\n",ecode);
      exit(2);
   }
   printf("main():dispositivo: [%s] tamao bloque: %d bytes\n",fs->device_name,fs->blocksize);
   printf("main():cantidad de i-nodos: %d cantidad de bloques %d\n",fs->super->s_inodes_count,fs->super->s_blocks_count);
   //if (!(fs->super->s_feature_compat & EXT3_FEATURE_COMPAT_HAS_JOURNAL)) printf("main():no tendria compatibilidad ext3?\n");
   //else printf("main():tendria compatibilidad ext3?\n");
   //printf("main():ext2fs_read_inode %d\n",ino);
   ecode =  ext2fs_read_inode (fs, ino, &inode);
   if ( ecode ) {
      printf("main():Error, ext2fs_read_inode() retorna: %ld\n",ecode);
      exit(3);
   }
   printf("main():i-nodo numero: %d tiene %d bloques asignados\nmain():Cantidad de punteros a bloques: %d:\n",ino,(int) inode.i_blocks,EXT2_N_BLOCKS);
   printf("main():imprimo punteros a bloques del i-nodo: %d\n",ino);
   int i;
   for(i=0;i<EXT2_N_BLOCKS;i++) {
      printf("\tmain():puntero numero: %2d apunta a bloque: %d\n",i,(int) inode.i_block[i]);
   }
   printf("main():el puntero de indireccion es: %d y apunta al bloque: %d\n",EXT2_IND_BLOCK,inode.i_block[EXT2_IND_BLOCK]);
   printf("main():el puntero de doble indireccion es: %d y apunta al bloque: %d\n",EXT2_DIND_BLOCK,inode.i_block[EXT2_DIND_BLOCK]);
   printf("main():el puntero de triple indireccion es: %d y apunta al bloque: %d\n",EXT2_TIND_BLOCK,inode.i_block[EXT2_TIND_BLOCK]);
   
   int tecla;
   if ( modo_interactivo ) { printf("\nmain():Intro continua mostrando contenido de bloques, q termina:");tecla = getchar(); }
   else tecla = '\n';
   if ( tecla != 'q' ) {
      printf("main():muestro bloques de datos del i-nodo: %d\n",ino);
      fs_data vars; // datos a ser pasados a la funcion de callback
      vars.contador=0;
      vars.total_bloques=(int) inode.i_blocks;
      vars.modo_interactivo = modo_interactivo;
      char *block_buffer =  (char *) malloc(fs->blocksize*3); // el buffer debe ser -al menos- 3 veces mas grande que el tamao del bloque del fs
      ecode = ext2fs_block_iterate (fs,ino,BLOCK_FLAG_DATA_ONLY|BLOCK_FLAG_READ_ONLY, block_buffer, leo_bloques_de_datos, &vars);
      free(block_buffer);
   }
   // simple,doble,triple indireccion
   imprimo_indireccion(1,fs,inode.i_block[EXT2_IND_BLOCK],modo_interactivo);
   imprimo_indireccion(2,fs,inode.i_block[EXT2_DIND_BLOCK],modo_interactivo);
   imprimo_indireccion(3,fs,inode.i_block[EXT2_TIND_BLOCK],modo_interactivo);
   //printf("main():ext2fs_close\n");
   ecode = ext2fs_close(fs);
   if ( ecode ) {
      printf("main():Error, ext2fs_close() retorna: %ld\n",ecode);
      exit(4);
   }
   //printf("main():fin\n");
   printf("main():fin\n");
   return 0;
}

int leo_bloques_de_datos(ext2_filsys fs, blk_t *blocknr, int blockcnt, void *private) {
   fs_data *vars = (fs_data *) private;
   vars->contador++;
   char block_buffer[fs->blocksize];
   errcode_t ecode =  leo_bloque(fs,*blocknr,block_buffer); 
   if ( ecode ) {
      printf("leo_bloques_de_datos():Error %ld leyendo bloque %d\n",ecode,*blocknr);
   } else {
      imprimo_bloque(block_buffer,fs->blocksize);
   }
   int tecla;
   if ( vars->modo_interactivo ) { printf("leo_bloques_de_datos():Imprimi bloque %d (%d de %d), Intro continua, q termina:",*blocknr,vars->contador,vars->total_bloques);tecla = getchar();}
   else { printf("leo_bloques_de_datos():Imprimi bloque %d (%d de %d)\n",*blocknr,vars->contador,vars->total_bloques);tecla='\n'; }
   if ( tecla == 'q' ) return BLOCK_ABORT;
   return 0; //(~BLOCK_CHANGED | ~BLOCK_ABORT);
}

// leo bloque fisico del file system (ya sea de data o metadata)
errcode_t leo_bloque(ext2_filsys fs,blk_t blocknr,char *block_buffer) {
   memset(block_buffer,0,fs->blocksize);
   errcode_t ecode =  io_channel_read_blk(fs->io, blocknr, -fs->blocksize,block_buffer);
   return ecode;
}

// muestro en pantalla bloque fisico del file system
void imprimo_bloque(char *block_buffer,int blocksize) {
   int i,w;
   for(i=0;i<blocksize;i+=16) {
      printf("%03X| ",i);
      for(w=0;w<16;w++) printf("%02X ",*(block_buffer+i+w));
      printf("|");
      for(w=0;w<16;w++) {
         if ( isprint(*(block_buffer+i+w)) ) printf("%c",*(block_buffer+i+w));
         else printf(".");
      }
      printf("\n");
   }
}

// interpreto bloque de indireccion como un arreglo de enteros
void imprimo_bloque_indireccion(char *block_buffer,int blocksize) {
   int *p = (int *) block_buffer;
   printf("imprimo_bloque_indireccion():bloque de indireccion apunta a los bloques:\n");
   while(*p && p < (int *)(block_buffer+blocksize)) { printf("%d ",*p);p++; }printf("\n");
}

//imprimo e interpreto bloque de indireccion (uso la misma funcion para los 3 niveles de indireccion)
void imprimo_indireccion(short indireccion,ext2_filsys fs,blk_t blocknr,short modo_interactivo) {
   char *tit[] = {"indireccion","doble indireccion","triple indireccion"};
   if ( blocknr ) {
      printf("imprimo_indireccion():muestro bloque de %s numero: %d\n",tit[indireccion-1],blocknr);
      char block_buffer[fs->blocksize];
      errcode_t ecode =  leo_bloque(fs,blocknr,block_buffer); 
      if ( ecode ) {
         printf("imprimo_indireccion():Error %ld leyendo bloque %d\n",ecode,blocknr);
      } else {
         imprimo_bloque(block_buffer,fs->blocksize);
         imprimo_bloque_indireccion(block_buffer,fs->blocksize);
         if ( modo_interactivo ) { printf("\nimprimo_indireccion():presione Intro para continuar:");getchar(); }
      }
   }
}